"""Link extractors using HTML5 parsers (selectolax/lexbor and markupever/html5ever).

This module provides link extraction functionality that works with HTML5-compliant
parsers instead of lxml. It integrates with Scrapy's link extraction framework while
supporting both the lexbor (via selectolax) and html5ever (via markupever) backends.

Example:
    >>> from scrapy_h5.extractor import LinkExtractor
    >>> from scrapy_h5 import HtmlFiveResponse
    >>> response = HtmlFiveResponse(url="http://example.com", body=b'<a href="/page">Link</a>')
    >>> extractor = LinkExtractor()
    >>> links = extractor.extract_links(response)
"""

import logging
import operator
from collections.abc import Callable, Iterable
from functools import partial
from typing import Any
from urllib.parse import urljoin

import markupever
import selectolax.lexbor
from parsel import Selector
from scrapy.link import Link
from scrapy.linkextractors.lxmlhtml import LxmlLinkExtractor, LxmlParserLinkExtractor, _nons, _RegexOrSeveral
from scrapy.utils.misc import rel_has_nofollow
from w3lib.html import strip_html5_whitespace
from w3lib.url import safe_url_string

logger = logging.getLogger(__name__)


class LinkExtractor(LxmlLinkExtractor):
    """A link extractor that uses HTML5 parsers for link extraction.

    This class extends Scrapy's LxmlLinkExtractor to work with HTML5-compliant
    parsers (selectolax/lexbor or markupever/html5ever) instead of lxml.

    Unlike the parent class, this extractor does not support `restrict_xpaths`,
    `restrict_css`, or `restrict_text` parameters as it operates directly on
    HTML5-parsed DOM trees.

    Args:
        allow: Regular expression(s) that URLs must match to be extracted.
        deny: Regular expression(s) that URLs must NOT match to be extracted.
        allow_domains: Domain(s) that URLs must belong to.
        deny_domains: Domain(s) that URLs must NOT belong to.
        tags: Tag name(s) to look for when extracting links. Defaults to ('a', 'area').
        attrs: Attribute name(s) to look for in tags. Defaults to ('href',).
        canonicalize: Whether to canonicalize URLs before filtering. Defaults to False.
        unique: Whether to deduplicate extracted links. Defaults to True.
        process_value: Optional callable to process each extracted URL value.
        deny_extensions: File extension(s) to deny. Uses default if None.
        strip: Whether to strip whitespace from URLs. Defaults to True.

    Example:
        >>> extractor = LinkExtractor(
        ...     allow=r"/products/",
        ...     deny=r"/admin/",
        ...     tags=["a"],
        ...     attrs=["href"],
        ... )
    """

    def __init__(  # noqa: PLR0913
        self,
        allow: _RegexOrSeveral = (),
        deny: _RegexOrSeveral = (),
        allow_domains: str | Iterable[str] = (),
        deny_domains: str | Iterable[str] = (),
        tags: str | Iterable[str] = ("a", "area"),
        attrs: str | Iterable[str] = ("href",),
        canonicalize: bool = False,  # noqa: FBT001, FBT002
        unique: bool = True,  # noqa: FBT001, FBT002
        process_value: Callable[[Any], Any] | None = None,
        deny_extensions: str | Iterable[str] | None = None,
        strip: bool = True,  # noqa: FBT001, FBT002
    ) -> None:
        super().__init__(
            allow=allow,
            deny=deny,
            allow_domains=allow_domains,
            deny_domains=deny_domains,
            restrict_xpaths=(),
            tags=tags,
            attrs=attrs,
            canonicalize=canonicalize,
            unique=unique,
            process_value=process_value,
            deny_extensions=deny_extensions,
            restrict_css=(),
            strip=strip,
            restrict_text=None,
        )
        self.link_extractor = HtmlFiveParserLinkExtractor(
            tag=partial(operator.contains, tags),
            attr=partial(operator.contains, attrs),
            unique=unique,
            process=process_value,
            strip=strip,
            canonicalized=not canonicalize,
        )


class HtmlFiveParserLinkExtractor(LxmlParserLinkExtractor):
    """Low-level link parser that extracts links from HTML5-parsed DOM trees.

    This class extends Scrapy's LxmlParserLinkExtractor to iterate over
    HTML5-parsed DOM nodes (from selectolax/lexbor or markupever/html5ever)
    instead of lxml etree elements.

    It overrides the `_iter_links` method to traverse the DOM tree and yield
    (element, attribute_name, attribute_value) tuples for matching tags/attrs.

    This class is used internally by LinkExtractor and typically should not
    be instantiated directly.
    """

    def _extract_links(
        self,
        selector: Selector,
        response_url: str,
        response_encoding: str,
        base_url: str,
    ) -> list[Link]:
        """Extract links from the selector's root document.

        This overrides the parent method to handle HTML5 element types
        (selectolax LexborNode and markupever Element) instead of lxml elements.

        Args:
            selector: A selector object with a `root` attribute pointing to the DOM tree.
            response_url: The URL of the response being processed.
            response_encoding: The character encoding of the response.
            base_url: The base URL for resolving relative links.

        Returns:
            A list of Link objects extracted from the document.
        """
        links: list[Link] = []
        for el, _attr, attr_val in self._iter_links(selector.root):
            try:
                if self.strip:
                    attr_val = strip_html5_whitespace(attr_val)  # noqa: PLW2901
                attr_val = urljoin(base_url, attr_val)  # noqa: PLW2901
            except ValueError:
                continue  # skipping bogus links
            else:
                url = self.process_attr(attr_val)
                if url is None:
                    continue
            try:
                url = safe_url_string(url, encoding=response_encoding)
            except ValueError:
                continue

            # to fix relative links after process_value
            url = urljoin(response_url, url)
            link = Link(
                url,
                self._get_element_text(el),
                nofollow=rel_has_nofollow(self._get_element_attr(el, "rel")),
            )
            links.append(link)
        return self._deduplicate_if_needed(links)

    def _get_element_text(
        self,
        el: selectolax.lexbor.LexborNode | markupever.dom.BaseNode,
    ) -> str:
        """Get the text content of an element.

        Args:
            el: The element to extract text from.

        Returns:
            The text content of the element, or empty string if none.
        """
        if isinstance(el, selectolax.lexbor.LexborNode):
            return el.text()
        if isinstance(el, markupever.dom.Element):
            return el.text()
        return ""

    def _get_element_attr(
        self,
        el: selectolax.lexbor.LexborNode | markupever.dom.BaseNode,
        attr: str,
    ) -> str | None:
        """Get an attribute value from an element.

        Args:
            el: The element to get the attribute from.
            attr: The attribute name.

        Returns:
            The attribute value, or None if not present.
        """
        if isinstance(el, selectolax.lexbor.LexborNode):
            return el.attributes.get(attr)
        if isinstance(el, markupever.dom.Element):
            return el.attrs.get(attr)
        return None

    def _iter_links(
        self,
        document: selectolax.lexbor.LexborNode | markupever.dom.BaseNode,
    ) -> Iterable[tuple[selectolax.lexbor.LexborNode | markupever.dom.BaseNode, str, str]]:
        """Iterate over links in the document.

        Dispatches to the appropriate backend-specific method based on document type.

        Args:
            document: The root node of the HTML5-parsed document tree.
                Can be either a selectolax LexborNode or markupever BaseNode.

        Yields:
            Tuples of (element, attribute_name, attribute_value) for each link found.

        Raises:
            TypeError: If document is not a supported node type.
        """
        if isinstance(document, selectolax.lexbor.LexborNode):
            yield from self._iter_links_lexbor(document)
        elif isinstance(document, markupever.dom.BaseNode):
            yield from self._iter_links_html5ever(document)
        else:
            raise TypeError(f"Unsupported document type: {type(document)}")

    def _iter_links_lexbor(
        self,
        document: selectolax.lexbor.LexborNode,
    ) -> Iterable[tuple[selectolax.lexbor.LexborNode, str, str]]:
        """Iterate over links in a selectolax/lexbor document.

        Traverses the DOM tree and yields link elements matching the configured
        tag and attribute filters.

        Args:
            document: The root LexborNode to traverse.

        Yields:
            Tuples of (element, attribute_name, attribute_value) for each link.
        """
        for el in document.traverse():
            if not self.scan_tag(_nons(el.tag)):
                continue
            attribs = el.attributes
            for attrib in attribs:
                if not self.scan_attr(attrib):
                    continue
                yield el, attrib, attribs[attrib]

    def _iter_links_html5ever(
        self,
        document: markupever.dom.BaseNode,
    ) -> Iterable[tuple[markupever.dom.BaseNode, str, str]]:
        """Iterate over links in a markupever/html5ever document.

        Traverses the DOM tree and yields link elements matching the configured
        tag and attribute filters.

        Args:
            document: The root BaseNode to traverse.

        Yields:
            Tuples of (element, attribute_name, attribute_value) for each link.
        """
        for edge in document.traverse():
            # Skip closed edges (only process opening traversal)
            if edge.closed:
                continue
            # EdgeTraverse wraps nodes - get the actual node
            el = edge.node
            if not isinstance(el, markupever.dom.Element):
                continue
            if not self.scan_tag(_nons(el.name.local)):
                continue
            attribs = el.attrs
            for attrib in attribs:
                if not self.scan_attr(attrib.local):
                    continue
                yield el, attrib, attribs[attrib]



"""Tests for LinkExtractor and HtmlFiveParserLinkExtractor."""

import pytest

from scrapy_h5 import HtmlFiveResponse, HtmlFiveSelector
from scrapy_h5.extractor import HtmlFiveParserLinkExtractor, LinkExtractor


class TestCssLinkExtractor:
    """Tests for LinkExtractor class."""

    SAMPLE_HTML = b"""
    <html>
    <head><title>Test Page</title></head>
    <body>
        <div id="nav">
            <a href="/home">Home</a>
            <a href="/about">About</a>
            <a href="/contact">Contact</a>
        </div>
        <div id="content">
            <h1>Welcome</h1>
            <p>Check out our <a href="/products">products</a>.</p>
            <a href="http://external.com/page">External Link</a>
        </div>
        <div id="footer">
            <a href="/privacy">Privacy Policy</a>
            <area href="/sitemap" alt="Sitemap">
        </div>
    </body>
    </html>
    """

    def _create_response(self, backend: str, html: bytes | None = None) -> HtmlFiveResponse:
        """Create a HtmlFiveResponse for testing."""
        body = html if html is not None else self.SAMPLE_HTML
        return HtmlFiveResponse(url="http://example.com", body=body).with_backend(backend)

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_all_links(self, backend: str) -> None:
        """Test extracting all links from HTML."""
        response = self._create_response(backend)
        extractor = LinkExtractor()
        links = extractor.extract_links(response)

        # Should find all <a> and <area> tags with href
        urls = [link.url for link in links]
        assert "/home" in urls or any("/home" in url for url in urls)
        assert "/about" in urls or any("/about" in url for url in urls)
        assert "/products" in urls or any("/products" in url for url in urls)

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_links_with_allow_pattern(self, backend: str) -> None:
        """Test link extraction with allow regex pattern."""
        response = self._create_response(backend)
        extractor = LinkExtractor(allow=r"/products")
        links = extractor.extract_links(response)

        urls = [link.url for link in links]
        # Only /products should match
        assert len(urls) == 1
        assert "/products" in urls[0]

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_links_with_deny_pattern(self, backend: str) -> None:
        """Test link extraction with deny regex pattern."""
        response = self._create_response(backend)
        extractor = LinkExtractor(deny=r"/privacy")
        links = extractor.extract_links(response)

        urls = [link.url for link in links]
        # /privacy should be excluded
        assert not any("/privacy" in url for url in urls)
        # But other links should still be present
        assert len(urls) >= 1

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_links_with_allow_domains(self, backend: str) -> None:
        """Test link extraction with domain filtering."""
        html = b"""
        <html><body>
            <a href="http://allowed.com/page1">Allowed 1</a>
            <a href="http://allowed.com/page2">Allowed 2</a>
            <a href="http://blocked.com/page">Blocked</a>
        </body></html>
        """
        response = self._create_response(backend, html)
        extractor = LinkExtractor(allow_domains=["allowed.com"])
        links = extractor.extract_links(response)

        urls = [link.url for link in links]
        assert all("allowed.com" in url for url in urls)
        assert not any("blocked.com" in url for url in urls)

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_links_with_deny_domains(self, backend: str) -> None:
        """Test link extraction with deny_domains."""
        html = b"""
        <html><body>
            <a href="http://good.com/page">Good</a>
            <a href="http://bad.com/page">Bad</a>
        </body></html>
        """
        response = self._create_response(backend, html)
        extractor = LinkExtractor(deny_domains=["bad.com"])
        links = extractor.extract_links(response)

        urls = [link.url for link in links]
        assert not any("bad.com" in url for url in urls)

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_links_custom_tags(self, backend: str) -> None:
        """Test link extraction with custom tags."""
        html = b"""
        <html><body>
            <a href="/link-a">A Tag</a>
            <link href="/link-link" rel="stylesheet">
            <script src="/script.js"></script>
        </body></html>
        """
        response = self._create_response(backend, html)
        extractor = LinkExtractor(tags=["link"], attrs=["href"])
        links = extractor.extract_links(response)

        urls = [link.url for link in links]
        assert any("/link-link" in url for url in urls)
        # <a> should not be extracted since we only specified 'link'
        assert not any("/link-a" in url for url in urls)

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_links_custom_attrs(self, backend: str) -> None:
        """Test link extraction with custom attributes."""
        html = b"""
        <html><body>
            <a href="/regular">Regular Link</a>
            <a data-url="/data-url">Data URL</a>
        </body></html>
        """
        response = self._create_response(backend, html)
        extractor = LinkExtractor(attrs=["data-url"])
        links = extractor.extract_links(response)

        urls = [link.url for link in links]
        assert any("/data-url" in url for url in urls)
        # href should not be extracted since we only specified 'data-url'
        assert not any("/regular" in url for url in urls)

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_links_unique(self, backend: str) -> None:
        """Test unique link filtering."""
        html = b"""
        <html><body>
            <a href="/same">Link 1</a>
            <a href="/same">Link 2</a>
            <a href="/same">Link 3</a>
        </body></html>
        """
        response = self._create_response(backend, html)

        # With unique=True (default)
        extractor_unique = LinkExtractor(unique=True)
        links_unique = extractor_unique.extract_links(response)
        assert len(links_unique) == 1

        # With unique=False
        extractor_not_unique = LinkExtractor(unique=False)
        links_not_unique = extractor_not_unique.extract_links(response)
        assert len(links_not_unique) == 3  # noqa: PLR2004

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_links_strip_whitespace(self, backend: str) -> None:
        """Test whitespace stripping from URLs."""
        html = b"""
        <html><body>
            <a href="  /spaced  ">Spaced Link</a>
        </body></html>
        """
        response = self._create_response(backend, html)

        # With strip=True (default)
        extractor = LinkExtractor(strip=True)
        links = extractor.extract_links(response)

        urls = [link.url for link in links]
        assert len(urls) == 1
        # URL should be stripped
        assert "  " not in urls[0]

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_links_deny_extensions(self, backend: str) -> None:
        """Test denying specific file extensions."""
        html = b"""
        <html><body>
            <a href="/page.html">HTML Page</a>
            <a href="/image.jpg">Image</a>
            <a href="/doc.pdf">PDF</a>
        </body></html>
        """
        response = self._create_response(backend, html)
        extractor = LinkExtractor(deny_extensions=["jpg", "pdf"])
        links = extractor.extract_links(response)

        urls = [link.url for link in links]
        assert any("/page.html" in url for url in urls)
        assert not any(".jpg" in url for url in urls)
        assert not any(".pdf" in url for url in urls)

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_links_empty_document(self, backend: str) -> None:
        """Test extracting links from empty document."""
        response = self._create_response(backend, b"<html><body></body></html>")
        extractor = LinkExtractor()
        links = extractor.extract_links(response)

        assert len(links) == 0

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_links_no_matching_tags(self, backend: str) -> None:
        """Test extraction when no tags match."""
        html = b"<html><body><div>No links here</div></body></html>"
        response = self._create_response(backend, html)
        extractor = LinkExtractor()
        links = extractor.extract_links(response)

        assert len(links) == 0

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_links_area_tag(self, backend: str) -> None:
        """Test that area tags are extracted by default."""
        html = b"""
        <html><body>
            <map name="imagemap">
                <area href="/area1" alt="Area 1">
                <area href="/area2" alt="Area 2">
            </map>
        </body></html>
        """
        response = self._create_response(backend, html)
        extractor = LinkExtractor()
        links = extractor.extract_links(response)

        urls = [link.url for link in links]
        assert any("/area1" in url for url in urls)
        assert any("/area2" in url for url in urls)

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_links_process_value(self, backend: str) -> None:
        """Test custom process_value callback."""
        html = b"""
        <html><body>
            <a href="/page?track=123">Link</a>
        </body></html>
        """
        response = self._create_response(backend, html)

        def remove_tracking(url: str) -> str:
            """Remove tracking parameters from URL."""
            if "?" in url:
                return url.split("?")[0]
            return url

        extractor = LinkExtractor(process_value=remove_tracking)
        links = extractor.extract_links(response)

        urls = [link.url for link in links]
        assert len(urls) == 1
        assert "track=" not in urls[0]
        assert "/page" in urls[0]

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_links_multiple_allow_patterns(self, backend: str) -> None:
        """Test multiple allow patterns."""
        html = b"""
        <html><body>
            <a href="/products/item1">Product 1</a>
            <a href="/services/service1">Service 1</a>
            <a href="/about">About</a>
        </body></html>
        """
        response = self._create_response(backend, html)
        extractor = LinkExtractor(allow=[r"/products", r"/services"])
        links = extractor.extract_links(response)

        urls = [link.url for link in links]
        assert len(urls) == 2  # noqa: PLR2004
        assert any("/products" in url for url in urls)
        assert any("/services" in url for url in urls)
        assert not any("/about" in url for url in urls)


class TestHtmlFiveParserLinkExtractor:
    """Tests for HtmlFiveParserLinkExtractor class."""

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_iter_links_basic(self, backend: str) -> None:
        """Test basic link iteration."""
        html = """
        <html><body>
            <a href="/link1">Link 1</a>
            <a href="/link2">Link 2</a>
        </body></html>
        """
        sel = HtmlFiveSelector(backend, text=html)
        extractor = HtmlFiveParserLinkExtractor(
            tag=lambda t: t == "a",
            attr=lambda a: a == "href",
        )

        links = list(extractor._iter_links(sel._root))  # noqa: SLF001
        assert len(links) == 2  # noqa: PLR2004

        # Each link should be (element, attr_name, attr_value)
        for _el, attr, value in links:
            assert attr == "href"
            assert value.startswith("/link")

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_iter_links_filters_tags(self, backend: str) -> None:
        """Test that tag filtering works."""
        html = """
        <html><body>
            <a href="/a-link">A Link</a>
            <area href="/area-link" alt="Area">
            <div href="/div-link">Div (should be ignored)</div>
        </body></html>
        """
        sel = HtmlFiveSelector(backend, text=html)
        extractor = HtmlFiveParserLinkExtractor(
            tag=lambda t: t == "a",
            attr=lambda a: a == "href",
        )

        links = list(extractor._iter_links(sel._root))  # noqa: SLF001
        values = [v for _, _, v in links]

        assert "/a-link" in values
        assert "/area-link" not in values
        assert "/div-link" not in values

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_iter_links_filters_attrs(self, backend: str) -> None:
        """Test that attribute filtering works."""
        html = """
        <html><body>
            <a href="/href-link">Href Link</a>
            <a data-link="/data-link">Data Link</a>
        </body></html>
        """
        sel = HtmlFiveSelector(backend, text=html)
        extractor = HtmlFiveParserLinkExtractor(
            tag=lambda t: t == "a",
            attr=lambda a: a == "data-link",
        )

        links = list(extractor._iter_links(sel._root))  # noqa: SLF001
        values = [v for _, _, v in links]

        assert "/data-link" in values
        assert "/href-link" not in values

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_iter_links_multiple_attrs(self, backend: str) -> None:
        """Test iteration with element having multiple matching attributes."""
        html = """
        <html><body>
            <a href="/href" data-link="/data">Multi Attr</a>
        </body></html>
        """
        sel = HtmlFiveSelector(backend, text=html)
        extractor = HtmlFiveParserLinkExtractor(
            tag=lambda t: t == "a",
            attr=lambda a: a in ("href", "data-link"),
        )

        links = list(extractor._iter_links(sel._root))  # noqa: SLF001
        values = [v for _, _, v in links]

        assert "/href" in values
        assert "/data" in values

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_iter_links_empty_document(self, backend: str) -> None:
        """Test iteration on empty document."""
        sel = HtmlFiveSelector(backend, text="<html><body></body></html>")
        extractor = HtmlFiveParserLinkExtractor(
            tag=lambda t: t == "a",
            attr=lambda a: a == "href",
        )

        links = list(extractor._iter_links(sel._root))  # noqa: SLF001
        assert len(links) == 0

    def test_iter_links_unsupported_type(self) -> None:
        """Test that unsupported document types raise TypeError."""
        extractor = HtmlFiveParserLinkExtractor(
            tag=lambda t: t == "a",
            attr=lambda a: a == "href",
        )

        with pytest.raises(TypeError, match="Unsupported document type"):
            list(extractor._iter_links("not a valid document"))  # noqa: SLF001

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_iter_links_nested_elements(self, backend: str) -> None:
        """Test iteration handles nested elements correctly."""
        html = """
        <html><body>
            <div>
                <nav>
                    <ul>
                        <li><a href="/link1">Link 1</a></li>
                        <li><a href="/link2">Link 2</a></li>
                    </ul>
                </nav>
            </div>
            <a href="/link3">Link 3</a>
        </body></html>
        """
        sel = HtmlFiveSelector(backend, text=html)
        extractor = HtmlFiveParserLinkExtractor(
            tag=lambda t: t == "a",
            attr=lambda a: a == "href",
        )

        links = list(extractor._iter_links(sel._root))  # noqa: SLF001
        values = [v for _, _, v in links]

        assert "/link1" in values
        assert "/link2" in values
        assert "/link3" in values
        assert len(links) == 3  # noqa: PLR2004

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_iter_links_with_empty_href(self, backend: str) -> None:
        """Test iteration handles empty href values."""
        html = """
        <html><body>
            <a href="">Empty Link</a>
            <a href="/valid">Valid Link</a>
        </body></html>
        """
        sel = HtmlFiveSelector(backend, text=html)
        extractor = HtmlFiveParserLinkExtractor(
            tag=lambda t: t == "a",
            attr=lambda a: a == "href",
        )

        links = list(extractor._iter_links(sel._root))  # noqa: SLF001
        values = [v for _, _, v in links]

        # Both should be yielded (filtering happens at higher level)
        assert "" in values
        assert "/valid" in values


class TestCssLinkExtractorInit:
    """Tests for LinkExtractor initialization."""

    def test_default_tags(self) -> None:
        """Test default tags are 'a' and 'area'."""
        extractor = LinkExtractor()
        # Check that the link_extractor has the right tag filter
        assert extractor.link_extractor.scan_tag("a")
        assert extractor.link_extractor.scan_tag("area")
        assert not extractor.link_extractor.scan_tag("div")

    def test_default_attrs(self) -> None:
        """Test default attrs is 'href'."""
        extractor = LinkExtractor()
        assert extractor.link_extractor.scan_attr("href")
        assert not extractor.link_extractor.scan_attr("src")

    def test_custom_tags(self) -> None:
        """Test custom tags parameter."""
        extractor = LinkExtractor(tags=["link", "script"])
        assert extractor.link_extractor.scan_tag("link")
        assert extractor.link_extractor.scan_tag("script")
        assert not extractor.link_extractor.scan_tag("a")

    def test_custom_attrs(self) -> None:
        """Test custom attrs parameter."""
        extractor = LinkExtractor(attrs=["src", "data-url"])
        assert extractor.link_extractor.scan_attr("src")
        assert extractor.link_extractor.scan_attr("data-url")
        assert not extractor.link_extractor.scan_attr("href")

    def test_link_extractor_type(self) -> None:
        """Test that link_extractor is HtmlFiveParserLinkExtractor."""
        extractor = LinkExtractor()
        assert isinstance(extractor.link_extractor, HtmlFiveParserLinkExtractor)



"""Tests for HtmlFiveSelector and HtmlFiveSelectorList."""

import pytest

from scrapy_h5 import HtmlFiveSelector, HtmlFiveSelectorList


class TestHtmlFiveSelector:
    """Tests for HtmlFiveSelector class."""

    SAMPLE_HTML = """
    <html>
    <head>
        <title>Test Page</title>
        <link rel="icon" href="/ico/favicon.ico" type="image/x-icon"/>
        <link rel="canonical" href="https://www.example.com/preferred-version-of-page" />
    </head>
    <body>
        <div id="main" class="container primary">
            <h1>Hello World</h1>
            <p class="intro">This is an introduction.</p>
            <ul>
                <li><a href="/link1">Link 1</a></li>
                <li><a href="/link2">Link 2</a></li>
                <li><a href="/link3">Link 3</a></li>
            </ul>
        </div>
        <div id="footer">
            <p>Footer text</p>
        </div>
    </body>
    </html>
    """

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_create_from_text(self, backend: str) -> None:
        """Test creating selector from HTML text."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        assert sel is not None

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_xpath(self, backend: str) -> None:
        """Test xpath() metod."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        try:
            sel.xpath("//li")
            raise AssertionError
        except NotImplementedError:
            pass

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_css_element(self, backend: str) -> None:
        """Test basic CSS element selector."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("h1")
        assert len(result) == 1
        assert "Hello World" in result.get()

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_css_id(self, backend: str) -> None:
        """Test CSS ID selector."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("#main")
        assert len(result) == 1
        assert "container" in result.get()

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_css_class(self, backend: str) -> None:
        """Test CSS class selector."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css(".intro")
        assert len(result) == 1
        assert "introduction" in result.get()

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_css_descendant(self, backend: str) -> None:
        """Test CSS descendant selector."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("ul a")
        assert len(result) == 3  # noqa: PLR2004

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_css_child(self, backend: str) -> None:
        """Test CSS child selector."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("ul > li")
        assert len(result) == 3  # noqa: PLR2004

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_css_attribute(self, backend: str) -> None:
        """Test CSS attribute selector."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("a[href]")
        assert len(result) == 3  # noqa: PLR2004

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_css_text_pseudo(self, backend: str) -> None:
        """Test ::text pseudo-element."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("h1::text")
        assert len(result) == 1
        assert result.get() == "Hello World"

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_css_attr_pseudo(self, backend: str) -> None:
        """Test ::attr() pseudo-element."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("a::attr(href)")
        assert len(result) == 3  # noqa: PLR2004
        hrefs = result.getall()
        assert "/link1" in hrefs
        assert "/link2" in hrefs
        assert "/link3" in hrefs

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_css_complex(self, backend: str) -> None:
        """Test CSS attribute selector."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("head > link[rel='canonical']::attr(href)")
        assert len(result) == 1
        assert result[0].get() == "https://www.example.com/preferred-version-of-page"

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_getall(self, backend: str) -> None:
        """Test getall() method."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("li a::text").getall()
        assert result == ["Link 1", "Link 2", "Link 3"]

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_get_with_default(self, backend: str) -> None:
        """Test get() with default value."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("nonexistent").get(default="not found")
        assert result == "not found"

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_attrib_property(self, backend: str) -> None:
        """Test attrib property."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        div = sel.css("#main")
        assert len(div) == 1
        assert div.attrib.get("id") == "main"
        assert "container" in div.attrib.get("class", "")

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_re(self, backend: str) -> None:
        """Test regex extraction."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("a::attr(href)").re(r"/link(\d+)")
        assert result == ["1", "2", "3"]

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_re_first(self, backend: str) -> None:
        """Test re_first() method."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("a::attr(href)").re_first(r"/link(\d+)")
        assert result == "1"

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_re_first_with_default(self, backend: str) -> None:
        """Test re_first() with default."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("a::attr(href)").re_first(r"notfound", default="default")
        assert result == "default"

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_bool_true(self, backend: str) -> None:
        """Test boolean conversion with content."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("h1")
        assert bool(result[0]) is True

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_bool_false(self, backend: str) -> None:
        """Test boolean conversion without content."""
        sel = HtmlFiveSelector(backend, text="<div></div>")
        result = sel.css("div::text")
        # With ::text, we get the element but its text is empty
        # The boolean of the selector should be False when text is empty
        if len(result) > 0:
            assert bool(result[0]) is False
        else:
            # No results is also acceptable
            assert len(result) == 0

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_repr(self, backend: str) -> None:
        """Test string representation."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("h1")
        repr_str = repr(result[0])
        assert "HtmlFiveSelector" in repr_str
        assert "h1" in repr_str

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_chained_selectors(self, backend: str) -> None:
        """Test chaining CSS selectors."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        # Chain element selectors
        result = sel.css("#main").css("a")
        assert len(result) == 3  # noqa: PLR2004
        # Then get text from the chain
        texts = result.css("::text").getall()
        assert len(texts) == 3  # noqa: PLR2004

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_drop_single_element(self, backend: str) -> None:
        """Test dropping a single element from the DOM."""
        sel = HtmlFiveSelector(backend, text="<div><span>remove me</span><p>keep me</p></div>")
        sel.css("span")[0].drop()
        result = sel.get()
        assert "<span>" not in result
        assert "remove me" not in result
        assert "<p>keep me</p>" in result

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_drop_on_text_node_is_noop(self, backend: str) -> None:
        """Test that drop() on a text node doesn't raise."""
        sel = HtmlFiveSelector(backend, text="<div>text</div>")
        text_sel = sel.css("div::text")[0]
        # Should not raise
        text_sel.drop()


class TestHtmlFiveSelectorList:
    """Tests for HtmlFiveSelectorList class."""

    SAMPLE_HTML = """
    <ul>
        <li><a href="/a">A</a></li>
        <li><a href="/b">B</a></li>
        <li><a href="/c">C</a></li>
    </ul>
    """

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_getitem_single(self, backend: str) -> None:
        """Test indexing single element."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("li")
        assert isinstance(result[0], HtmlFiveSelector)

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_getitem_slice(self, backend: str) -> None:
        """Test slicing."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("li")
        sliced = result[1:]
        assert isinstance(sliced, HtmlFiveSelectorList)
        assert len(sliced) == 2  # noqa: PLR2004

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_css_on_list(self, backend: str) -> None:
        """Test CSS on SelectorList."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("li").css("a::text")
        assert result.getall() == ["A", "B", "C"]

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_attrib_on_list(self, backend: str) -> None:
        """Test attrib property on list."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("a")
        # Returns first element's attrib
        assert result.attrib.get("href") == "/a"

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_attrib_on_empty_list(self, backend: str) -> None:
        """Test attrib on empty list."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("nonexistent")
        assert result.attrib == {}

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_alias(self, backend: str) -> None:
        """Test extract() is alias for getall()."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("a::text")
        assert result.extract() == result.getall()

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_extract_first_alias(self, backend: str) -> None:
        """Test extract_first() is alias for get()."""
        sel = HtmlFiveSelector(backend, text=self.SAMPLE_HTML)
        result = sel.css("a::text")
        assert result.extract_first() == result.get()

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_drop_multiple_elements(self, backend: str) -> None:
        """Test dropping multiple elements from the DOM."""
        sel = HtmlFiveSelector(backend, text="<div><span>1</span><span>2</span><p>keep</p></div>")
        sel.css("span").drop()
        result = sel.get()
        assert "<span>" not in result
        assert "<p>keep</p>" in result

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_drop_nested_element(self, backend: str) -> None:
        """Test dropping a nested element."""
        sel = HtmlFiveSelector(backend, text="<div><ul><li>item</li></ul><p>text</p></div>")
        sel.css("ul").drop()
        result = sel.get()
        assert "<ul>" not in result
        assert "<li>" not in result
        assert "<p>text</p>" in result

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_drop_on_empty_list_is_noop(self, backend: str) -> None:
        """Test that drop() on empty SelectorList doesn't raise."""
        sel = HtmlFiveSelector(backend, text="<div>text</div>")
        empty = sel.css("nonexistent")
        # Should not raise
        empty.drop()



"""HtmlFiveResponse class extending Scrapy's HtmlResponse with html5 parsing."""

import logging
from typing import Any

from scrapy.http import HtmlResponse

from scrapy_h5.selector import HtmlFiveSelector, HtmlFiveSelectorList

logger = logging.getLogger(__name__)


class HtmlFiveResponse(HtmlResponse):
    """HtmlResponse subclass that uses html5 for parsing.

    This response class provides the same API as Scrapy's HtmlResponse,
    but uses html5 parser instead of lxml, providing better
    HTML5 compliance.
    """

    _html5_backend = None
    _cached_h5_selector: HtmlFiveSelector | None = None

    def with_backend(self, backend: str) -> "HtmlFiveResponse":
        self._html5_backend = backend
        return self

    @property
    def selector(self) -> HtmlFiveSelector:
        """Return an HtmlFiveSelector for this response's content.

        The selector is lazily created and cached.
        """
        if self._cached_h5_selector is None:
            self._cached_h5_selector = HtmlFiveSelector(self._html5_backend, text=self.text)
        return self._cached_h5_selector

    def xpath(
        self,
        query: str,
        **kwargs: Any,  # noqa: ANN401
    ) -> HtmlFiveSelectorList:
        """Select elements using XPath (converted to CSS).

        Note: Only common XPath patterns are supported. Complex expressions
        will raise XPathConversionError.
        """
        return self.selector.xpath(query, **kwargs)

    def css(self, query: str) -> HtmlFiveSelectorList:
        """Select elements using CSS selectors.

        Supports standard CSS selectors plus parsel's ::text and ::attr()
        pseudo-elements.
        """
        return self.selector.css(query)



"""HtmlFiveSelector and HtmlFiveSelectorList classes for HTML parsing."""

import logging
import re
from collections.abc import Mapping
from re import Pattern
from typing import Any, TypeVar, overload

import markupever
import selectolax.lexbor
from parsel.utils import extract_regex, flatten, iflatten, shorten
from scrapy.utils.trackref import object_ref

_SelectorType = TypeVar("_SelectorType", bound="HtmlFiveSelector")
_SelectorListType = TypeVar("_SelectorListType", bound="HtmlFiveSelectorList")

logger = logging.getLogger(__name__)


class HtmlFiveSelectorList(list[_SelectorType], object_ref):
    """A list of HtmlFiveSelector objects with convenience methods for bulk operations."""

    @overload
    def __getitem__(self, pos: int) -> _SelectorType: ...

    @overload
    def __getitem__(self, pos: slice) -> _SelectorListType: ...

    def __getitem__(self, pos: int | slice) -> _SelectorListType | _SelectorType:
        o = super().__getitem__(pos)
        if isinstance(pos, slice):
            return self.__class__(o)
        return o

    def __getstate__(self) -> None:
        raise TypeError("can't pickle SelectorList objects")

    def jmespath(
        self,
        query: str,
        **kwargs: Any,  # noqa: ANN401
    ) -> _SelectorListType:
        """Call the `.jmespath()` method for each element in this list."""
        return self.__class__(flatten([x.jmespath(query, **kwargs) for x in self]))

    def xpath(
        self,
        query: str,
        **kwargs: Any,  # noqa: ANN401
    ) -> _SelectorListType:
        """Apply XPath query (converted to CSS) to all elements and return flattened results."""
        return self.__class__(flatten([x.xpath(query, **kwargs) for x in self]))

    def css(self, query: str) -> _SelectorListType:
        """Apply CSS selector to all elements and return flattened results."""
        return self.__class__(flatten([x.css(query) for x in self]))

    def re(
        self,
        regex: str | Pattern[str],
        replace_entities: bool = True,  # noqa: FBT001, FBT002
    ) -> list[str]:
        """Apply regex to all elements and return flattened string results."""
        return flatten([x.re(regex, replace_entities=replace_entities) for x in self])

    @overload
    def re_first(
        self,
        regex: str | Pattern[str],
        default: None = None,
        replace_entities: bool = True,  # noqa: FBT001, FBT002
    ) -> str | None: ...

    @overload
    def re_first(
        self,
        regex: str | Pattern[str],
        default: str,
        replace_entities: bool = True,  # noqa: FBT001, FBT002
    ) -> str: ...

    def re_first(
        self,
        regex: str | Pattern[str],
        default: str | None = None,
        replace_entities: bool = True,  # noqa: FBT001, FBT002
    ) -> str | None:
        """Apply regex and return first match, or default if no match."""
        for el in iflatten(x.re(regex, replace_entities=replace_entities) for x in self):
            return el
        return default

    def getall(self) -> list[str]:
        """Get serialized content of all elements."""
        return [x.get() for x in self]

    extract = getall

    @overload
    def get(self, default: None = None) -> str | None: ...

    @overload
    def get(self, default: str) -> str: ...

    def get(self, default: str | None = None) -> str | None:
        """Get serialized content of first element, or default if empty."""
        for x in self:
            return x.get()
        return default

    extract_first = get

    @property
    def attrib(self) -> Mapping[str, str]:
        """Return attributes of first element, or empty dict if empty."""
        for x in self:
            return x.attrib
        return {}

    def drop(self) -> None:
        """Drop matched nodes from the parent for each element in this list."""
        for x in self:
            x.drop()


class HtmlFiveSelector(object_ref):
    """Selector for HTML parsing.

    Provides a parsel-compatible API for CSS and XPath selectors.
    """

    __slots__ = ["__weakref__", "_attr", "_backend", "_expr", "_root", "_text"]

    selectorlist_cls = HtmlFiveSelectorList

    @property
    def root(self) -> selectolax.lexbor.LexborNode | markupever.dom.BaseNode:
        """Return the root node of the selector (compatibility with Scrapy's Selector API)."""
        return self._root

    def __init__(
        self,
        backend: str,
        *,
        text: str | None = None,
        root: selectolax.lexbor.LexborNode | markupever.dom.BaseNode | None = None,
        _expr: str | None = None,
        _text: bool = False,
        _attr: str | None = None,
    ) -> None:
        if backend not in {"lexbor", "html5ever"}:
            raise ValueError(f"Unsupported html5 backend: {backend}")
        self._backend = backend

        if text is None and root is None:
            raise ValueError("At least one of text or root arguments must be provided")
        if text is not None and root is not None:
            raise ValueError("At most one of text or root arguments must be provided")

        if text is not None and not isinstance(text, str):
            raise TypeError(f"Argument `text` should be of type str, got {type(text)}")

        self._expr = _expr
        self._text = _text
        self._attr = _attr

        if text is not None and self._backend == "lexbor":
            # Parse the HTML text with Lexbor
            self._root = selectolax.lexbor.LexborHTMLParser(text).root
        elif text is not None and self._backend == "html5ever":
            # Parse the HTML text with html5ever
            self._root = markupever.parse(text).root()
        else:  # root is not None
            self._root = root

    def jmespath(
        self,
        query: str,
        **kwargs: Any,  # noqa: ANN401
    ) -> _SelectorListType:
        """Find objects matching the JMESPath `query`."""
        raise NotImplementedError

    def xpath(
        self,
        query: str,
        **kwargs: Any,  # noqa: ANN401
    ) -> _SelectorListType:
        """Select elements using XPath."""
        raise NotImplementedError

    def css(self, query: str) -> _SelectorListType:
        """Select elements using CSS selector.

        Supports parsel's ::text and ::attr() pseudo-elements.
        """
        results = self.selectorlist_cls()
        for subquery, is_text, attr_name in self._parse_css(query):
            results.extend(self._select_css(subquery, is_text, attr_name))

        return results

    def _parse_css(self, query: str) -> list[tuple[str, bool, str | None]]:
        """Parse CSS query and extract ::text and ::attr() pseudo-elements.

        Handles comma-separated selectors by processing each part.
        Returns: list of (subquery, is_text, attr_name)
        """
        # Handle comma-separated selectors
        if "," in query:
            parts = []
            for part in query.split(","):
                parts.extend(self._parse_css(part.strip()))
            return parts

        is_text = False
        attr_name = None

        # Check for ::text pseudo-element
        if query.endswith("::text"):
            query = query[:-6]
            is_text = True
        # Check for ::attr(name) pseudo-element
        elif match := re.search(r"::attr\(([^)]+)\)$", query):
            query = query[: match.start()]
            attr_name = match.group(1)

        return [(query, is_text, attr_name)]

    def _select_css(self, query: str, is_text: bool, attr_name: str | None) -> list[_SelectorType]:  # noqa: FBT001
        if not query.strip():
            # Apply pseudo-element to current element
            return [
                self.__class__(
                    self._backend,
                    root=self._root,
                    _expr=query,
                    _text=is_text,
                    _attr=attr_name,
                ),
            ]

        elements = (
            self._root.css(query) if isinstance(self._root, selectolax.lexbor.LexborNode) else self._root.select(query)
        )
        return [
            self.__class__(
                self._backend,
                root=el,
                _expr=query,
                _text=is_text,
                _attr=attr_name,
            )
            for el in elements
        ]

    def re(
        self,
        regex: str | Pattern[str],
        replace_entities: bool = True,  # noqa: FBT001, FBT002
    ) -> list[str]:
        """Apply regex to the selector content."""
        return extract_regex(regex, self.get(), replace_entities=replace_entities)

    @overload
    def re_first(
        self,
        regex: str | Pattern[str],
        default: None = None,
        replace_entities: bool = True,  # noqa: FBT001, FBT002
    ) -> str | None: ...

    @overload
    def re_first(
        self,
        regex: str | Pattern[str],
        default: str,
        replace_entities: bool = True,  # noqa: FBT001, FBT002
    ) -> str: ...

    def re_first(
        self,
        regex: str | Pattern[str],
        default: str | None = None,
        replace_entities: bool = True,  # noqa: FBT001, FBT002
    ) -> str | None:
        """Apply regex and return first match."""
        return next(
            iflatten(self.re(regex, replace_entities=replace_entities)),
            default,
        )

    def get(self) -> str:  # noqa: PLR0911
        """Serialize and return the matched node content."""
        if self._text and isinstance(self._root, selectolax.lexbor.LexborNode):
            return self._root.text()
        if self._text and isinstance(self._root, markupever.dom.Element):
            return self._root.text()
        if self._text:
            return ""

        if self._attr and isinstance(self._root, selectolax.lexbor.LexborNode):
            return self._root.attributes.get(self._attr, "")
        if self._attr and isinstance(self._root, markupever.dom.Element):
            return self._root.attrs.get(self._attr, "")
        if self._attr:
            return ""

        if isinstance(self._root, selectolax.lexbor.LexborNode):
            return self._root.html
        if isinstance(self._root, markupever.dom.BaseNode):
            return self._root.serialize()

        return str(self._root)

    extract = get

    def getall(self) -> list[str]:
        """Return the result in a 1-element list."""
        return [self.get()]

    def register_namespace(self, prefix: str, uri: str) -> None:
        """Register a namespace (kept for API compatibility)."""

    def remove_namespaces(self) -> None:
        """Remove namespaces (kept for API compatibility)."""

    def drop(self) -> None:
        """Drop/remove matched nodes from the parent.

        Removes the current element from its parent in the DOM tree.
        After calling drop(), the element is detached and the parent's
        serialized content will no longer include this element.
        """
        if isinstance(self._root, selectolax.lexbor.LexborNode):
            self._root.decompose()
        if isinstance(self._root, markupever.dom.BaseNode):
            self._root.detach()

    @property
    def attrib(self) -> dict[str, str | None]:
        """Return element attributes as a dict."""
        if isinstance(self._root, selectolax.lexbor.LexborNode):
            return self._root.attributes
        if isinstance(self._root, markupever.dom.Element):
            # Convert AttrsList to dict with proper key extraction
            # markupever uses QualName objects as keys, we need the local name
            attrs = self._root.attrs
            return {key.local: attrs.get(key, "") for key in attrs}
        return {}

    def __bool__(self) -> bool:
        """Return True if there is content."""
        return bool(self.get())

    __nonzero__ = __bool__

    def __str__(self) -> str:
        return self.get()

    def __repr__(self) -> str:
        data = repr(shorten(self.get(), width=40))
        return f"<{type(self).__name__} query={self._expr!r} data={data}>"



"""Scrapy Downloader Middleware for html5-based HTML parsing."""

import logging

from scrapy import Request, Spider, signals
from scrapy.crawler import Crawler
from scrapy.http import HtmlResponse, Response

from scrapy_h5.response import HtmlFiveResponse

logger = logging.getLogger(__name__)


class HtmlFiveResponseMiddleware:
    """Downloader Middleware that replaces HtmlResponse with HtmlFiveResponse.

    This middleware intercepts HTML responses and replaces them with
    HtmlFiveResponse instances, which use html5 for parsing instead of lxml.

    Settings:
        HTML5_BACKEND: str | None (default: 'lexbor')
            Global enable/disable for html5 parsing.

    Per-request control:
        Set request.meta['use_html5'] = False to disable for a specific request.
        Set request.meta['use_html5'] = 'html5ever' to force enable (overrides HTML5_BACKEND=False).

    Usage in settings.py:
        DOWNLOADER_MIDDLEWARES = {
            'scrapy_h5.HtmlFiveResponseMiddleware': 543,
        }
        HTML5_BACKEND = 'html5ever'  # optional, 'lexbor' by default
    """

    def __init__(self, *, backend: str | None = "lexbor") -> None:
        if backend not in {"lexbor", "html5ever", None}:
            raise ValueError(f"Unsupported html5 backend: {backend}")
        self.backend = backend

    @classmethod
    def from_crawler(cls, crawler: Crawler) -> "HtmlFiveResponseMiddleware":
        """Create middleware instance from crawler settings."""
        backend = crawler.settings.get("HTML5_BACKEND", default="lexbor")
        middleware = cls(backend=backend)
        crawler.signals.connect(middleware.spider_opened, signal=signals.spider_opened)
        return middleware

    def spider_opened(self, spider: Spider) -> None:
        """Log when spider opens."""
        spider.logger.debug(
            "HtmlFiveResponseMiddleware backend=%s (html5 parsing for HTML responses)",
            self.backend,
        )

    def process_response(
        self,
        request: Request,
        response: Response,
    ) -> Response:
        """Process response and optionally replace with HtmlFiveResponse.

        Only HtmlResponse instances are replaced. Other response types
        (XmlResponse, JsonResponse, TextResponse, etc.) pass through unchanged.
        """
        # Check if this is an HTML response
        if not isinstance(response, HtmlResponse):
            return response

        # Check per-request override
        use_html5 = request.meta.get("use_html5", self.backend)
        if use_html5 is None:
            # Disabled globally or explicitly (for this request)
            return response

        # Replace with HtmlFiveResponse
        return response.replace(cls=HtmlFiveResponse).with_backend(use_html5)



"""Tests for HtmlFiveResponseMiddleware."""

from unittest.mock import MagicMock

import pytest
from scrapy import Request, Spider
from scrapy.http import HtmlResponse, JsonResponse, TextResponse, XmlResponse

from scrapy_h5 import HtmlFiveResponse, HtmlFiveResponseMiddleware, HtmlFiveSelector


class TestHtmlFiveResponseMiddleware:
    """Tests for HtmlFiveResponseMiddleware class."""

    def _create_middleware(self, *, backend: str | None) -> HtmlFiveResponseMiddleware:
        """Create a middleware instance."""
        return HtmlFiveResponseMiddleware(backend=backend)

    def _create_request(self, url: str = "http://example.com", **meta: object) -> Request:
        """Create a test request."""
        return Request(url=url, meta=meta)

    def _create_html_response(
        self,
        url: str = "http://example.com",
        body: bytes = b"<html><body>Test</body></html>",
    ) -> HtmlResponse:
        """Create a test HTML response."""
        return HtmlResponse(url=url, body=body)

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_process_html_response_enabled(self, backend: str) -> None:
        """Test that HTML responses are replaced when backend."""
        middleware = self._create_middleware(backend=backend)
        request = self._create_request()
        response = self._create_html_response()

        result = middleware.process_response(request, response)

        assert isinstance(result, HtmlFiveResponse)
        assert result.url == response.url
        assert result.body == response.body

    def test_process_html_response_disabled(self) -> None:
        """Test that HTML responses are not replaced when disabled."""
        middleware = self._create_middleware(backend=None)
        request = self._create_request()
        response = self._create_html_response()

        result = middleware.process_response(request, response)

        assert isinstance(result, HtmlResponse)
        assert not isinstance(result, HtmlFiveResponse)

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_process_response_meta_disable(self, backend: str) -> None:
        """Test per-request disable via meta."""
        middleware = self._create_middleware(backend=backend)
        request = self._create_request(use_html5=None)
        response = self._create_html_response()

        result = middleware.process_response(request, response)

        assert isinstance(result, HtmlResponse)
        assert not isinstance(result, HtmlFiveResponse)

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_process_response_meta_enable_override(self, backend: str) -> None:
        """Test per-request enable overrides global disable."""
        middleware = self._create_middleware(backend=None)
        request = self._create_request(use_html5=backend)
        response = self._create_html_response()

        result = middleware.process_response(request, response)

        assert isinstance(result, HtmlFiveResponse)

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_xml_response_passthrough(self, backend: str) -> None:
        """Test that XML responses pass through unchanged."""
        middleware = self._create_middleware(backend=backend)
        request = self._create_request()
        response = XmlResponse(url="http://example.com", body=b"<root/>")

        result = middleware.process_response(request, response)

        assert isinstance(result, XmlResponse)
        assert result is response

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_json_response_passthrough(self, backend: str) -> None:
        """Test that JSON responses pass through unchanged."""
        middleware = self._create_middleware(backend=backend)
        request = self._create_request()
        response = JsonResponse(url="http://example.com", body=b'{"key": "value"}')

        result = middleware.process_response(request, response)

        assert isinstance(result, JsonResponse)
        assert result is response

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_text_response_passthrough(self, backend: str) -> None:
        """Test that plain text responses pass through unchanged."""
        middleware = self._create_middleware(backend=backend)
        request = self._create_request()
        response = TextResponse(url="http://example.com", body=b"plain text")

        result = middleware.process_response(request, response)

        # TextResponse is not HtmlResponse, so passes through
        assert isinstance(result, TextResponse)
        assert not isinstance(result, HtmlResponse)

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_already_h5_response_passthrough(self, backend: str) -> None:
        """Test that HtmlFiveResponse passes through unchanged."""
        middleware = self._create_middleware(backend=backend)
        request = self._create_request()
        response = HtmlFiveResponse(url="http://example.com", body=b"<html>test</html>")

        result = middleware.process_response(request, response)

        assert isinstance(result, HtmlFiveResponse)
        assert result.url == response.url
        assert result.body == response.body

    def test_from_crawler(self) -> None:
        """Test from_crawler class method."""
        crawler = MagicMock()
        crawler.settings.get.return_value = "html5ever"
        crawler.signals = MagicMock()

        middleware = HtmlFiveResponseMiddleware.from_crawler(crawler)

        assert middleware.backend == "html5ever"
        crawler.settings.get.assert_called_once_with("HTML5_BACKEND", default="lexbor")

    def test_from_crawler_disabled(self) -> None:
        """Test from_crawler with HTML5_BACKEND=False."""
        crawler = MagicMock()
        crawler.settings.get.return_value = "html5ever"
        crawler.signals = MagicMock()

        middleware = HtmlFiveResponseMiddleware.from_crawler(crawler)

        assert middleware.backend == "html5ever"

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_spider_opened_logging(self, backend: str) -> None:
        """Test spider_opened logs correctly."""
        middleware = self._create_middleware(backend=backend)
        spider = MagicMock(spec=Spider)

        # Should not raise
        middleware.spider_opened(spider)

        spider.logger.debug.assert_called_once()
        call_args = spider.logger.debug.call_args
        assert "HtmlFiveResponseMiddleware" in call_args[0][0]
        assert call_args[0][1] is backend


class TestHtmlFiveResponse:
    """Tests for HtmlFiveResponse class."""

    SAMPLE_HTML = b"""
    <html>
    <body>
        <h1>Title</h1>
        <p class="content">Paragraph</p>
        <a href="/link">Link</a>
    </body>
    </html>
    """

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_selector_property(self, backend: str) -> None:
        """Test selector property returns HtmlFiveSelector."""
        response = HtmlFiveResponse(url="http://example.com", body=self.SAMPLE_HTML).with_backend(backend)
        assert isinstance(response.selector, HtmlFiveSelector)

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_selector_cached(self, backend: str) -> None:
        """Test selector is cached."""
        response = HtmlFiveResponse(url="http://example.com", body=self.SAMPLE_HTML).with_backend(backend)
        selector1 = response.selector
        selector2 = response.selector
        assert selector1 is selector2

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_css_method(self, backend: str) -> None:
        """Test css() method."""
        response = HtmlFiveResponse(url="http://example.com", body=self.SAMPLE_HTML).with_backend(backend)
        result = response.css("h1::text").get()
        assert result == "Title"

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_xpath_method(self, backend: str) -> None:
        """Test xpath() method."""
        response = HtmlFiveResponse(url="http://example.com", body=self.SAMPLE_HTML).with_backend(backend)
        try:
            response.xpath("//a/@href").get()
            raise AssertionError
        except NotImplementedError:
            pass

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_css_getall(self, backend: str) -> None:
        """Test css().getall()."""
        response = HtmlFiveResponse(url="http://example.com", body=self.SAMPLE_HTML).with_backend(backend)
        result = response.css("h1::text, p::text").getall()
        # Note: CSS selector order may vary
        assert "Title" in result

    @pytest.mark.parametrize("backend", ["lexbor", "html5ever"])
    def test_complex_selectors(self, backend: str) -> None:
        """Test complex CSS selectors work."""
        response = HtmlFiveResponse(url="http://example.com", body=self.SAMPLE_HTML).with_backend(backend)
        result = response.css("p.content::text").get()
        assert result == "Paragraph"



"""scrapy-h5: HTML5 parsing for Scrapy.

This package provides Scrapy integration for html5-based HTML parsing.

Main components:
- HtmlFiveSelector: Selector class using html5 for parsing
- HtmlFiveSelectorList: List of HtmlFiveSelector instances
- HtmlFiveResponse: Scrapy response class with HtmlFiveSelector
- HtmlFiveResponseMiddleware: Scrapy Downloader Middleware for automatic html5 parsing
- LinkExtractor: Link extractor using HTML5 parsers

Usage:
    # In settings.py
    DOWNLOADER_MIDDLEWARES = {
        'scrapy_h5.HtmlFiveResponseMiddleware': 543,
    }

    # In spider
    def parse(self, response):
        # CSS selectors work directly
        titles = response.css('h1::text').getall()

        # Common XPath patterns are converted to CSS
        links = response.xpath('//a/@href').getall()

    # Using LinkExtractor
    from scrapy_h5 import LinkExtractor
    extractor = LinkExtractor(allow=r'/products/')
    links = extractor.extract_links(response)
"""

from scrapy_h5.extractor import LinkExtractor
from scrapy_h5.middleware import HtmlFiveResponseMiddleware
from scrapy_h5.response import HtmlFiveResponse
from scrapy_h5.selector import (
    HtmlFiveSelector,
    HtmlFiveSelectorList,
)

__all__ = [
    "HtmlFiveResponse",
    "HtmlFiveResponseMiddleware",
    "HtmlFiveSelector",
    "HtmlFiveSelectorList",
    "LinkExtractor",
]

__version__ = "0.1.0"
